extern "C"
{
  #include <wdm.h>
};
#include "prog.h"
#include "devio.h"

#define DEVICE	L"WINRING"
//===============================================
class RingDriver
{
public:
	static WCHAR DosDeviceName[];
	static WCHAR DeviceName[];

	static VOID Unload(IN PDRIVER_OBJECT DriverObject);
	static NTSTATUS Create(IN PDEVICE_OBJECT fdo, IN PIRP Irp);
	static NTSTATUS Close(IN PDEVICE_OBJECT fdo, IN PIRP Irp);
	static NTSTATUS DeviceControl(IN PDEVICE_OBJECT fdo, IN PIRP Irp);
	static NTSTATUS CompleteIrp(PIRP Irp, NTSTATUS status, ULONG info);
	static NTSTATUS AdjRing(PIRP);
	static void ToggleAPP(TRingData *p);
};

//===============================================
WCHAR RingDriver::DosDeviceName[]=L"\\DosDevices\\" DEVICE;
WCHAR RingDriver::DeviceName[]=L"\\Device\\" DEVICE;

//===============================================
extern "C"
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath)
{
	NTSTATUS status;
	UNICODE_STRING uniNtNameString;
	UNICODE_STRING uniWin32NameString;
	PDEVICE_OBJECT deviceObject;

	RtlInitUnicodeString( &uniNtNameString, RingDriver::DeviceName);

	status = IoCreateDevice(
                 DriverObject,
                 0,                    // We don't use a device extension
                 &uniNtNameString,
                 FILE_DEVICE_UNKNOWN,
                 0,                    // No standard device characteristics
                 FALSE,                // This isn't an exclusive device
                 &deviceObject
                 );

	if (!NT_SUCCESS(status))
		return status;

	DriverObject->MajorFunction[IRP_MJ_CREATE] = RingDriver::Create;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = RingDriver::DeviceControl;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = RingDriver::Close;
	DriverObject->DriverUnload = RingDriver::Unload;

	RtlInitUnicodeString( &uniWin32NameString, RingDriver::DosDeviceName);
	status = IoCreateSymbolicLink( &uniWin32NameString, &uniNtNameString );
	if (!NT_SUCCESS(status))
	  return status;

	return status;
}

//==============================================
VOID RingDriver::Unload(IN PDRIVER_OBJECT DriverObject)
{
	UNICODE_STRING uniWin32NameString;

	RtlInitUnicodeString( &uniWin32NameString, DosDeviceName);
	IoDeleteSymbolicLink( &uniWin32NameString );
	IoDeleteDevice( DriverObject->DeviceObject );
}

//==============================================
NTSTATUS RingDriver::Create(IN PDEVICE_OBJECT fdo, IN PIRP Irp)
{
	return CompleteIrp(Irp,STATUS_SUCCESS,0);
}

//==============================================
NTSTATUS RingDriver::Close(IN PDEVICE_OBJECT fdo, IN PIRP Irp)
{
	return CompleteIrp(Irp,STATUS_SUCCESS,0);
}

//==============================================
NTSTATUS RingDriver::DeviceControl(IN PDEVICE_OBJECT fdo, IN PIRP irp)
{
	PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(irp);
	NTSTATUS status = STATUS_SUCCESS;
	ULONG BytesTxd = 0;

	ULONG ControlCode = irpStack->Parameters.DeviceIoControl.IoControlCode;
	ULONG InputLength = irpStack->Parameters.DeviceIoControl.InputBufferLength;
	ULONG OutputLength = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
	switch( ControlCode)
	{
	  case WINRING_Access:
	    status = AdjRing(irp);
	    return status;
	  default:
	   status = STATUS_INVALID_DEVICE_REQUEST;
	}
	return CompleteIrp(irp,status,BytesTxd);
}

//===============================================
NTSTATUS RingDriver::CompleteIrp(PIRP Irp, NTSTATUS status, ULONG info)
{
	Irp->IoStatus.Status = status;
	Irp->IoStatus.Information = info;
	IoCompleteRequest(Irp,IO_NO_INCREMENT);
	return status;
}

//===============================================
NTSTATUS RingDriver::AdjRing(PIRP irp)
{
	PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation( irp );
	ULONG dwInputLength = irpStack->Parameters.DeviceIoControl.InputBufferLength;
	ULONG dwOutputLength= irpStack->Parameters.DeviceIoControl.OutputBufferLength;
	if (dwInputLength!=dwOutputLength && dwInputLength<sizeof(TRingData))
	{
		return CompleteIrp(irp,STATUS_DEVICE_DATA_ERROR,0);
	}

	TRingData *p=(TRingData *)irp->AssociatedIrp.SystemBuffer;

	ToggleAPP(p);

	return CompleteIrp(irp,0,dwInputLength);
}

//===============================================
// Run User's Procedure Function
//===============================================
void RingDriver::ToggleAPP(TRingData *p)
{
	TRingData Ring=*p;
	ULONG *pt;
	pt=Ring.RegData;

        _asm {
	  pushad
          mov eax, Ring.AdjRing0Entry		// Get the User's procedure of Application
	  push eax	
	  mov eax, pt
	  mov ebx, [eax][4]			// restore EBX
	  mov ecx, [eax][8]			// restore ECX
	  mov edx, [eax][12]			// restore EDX	
	  mov esi, [eax][16]			// restore ESI
	  mov edi, [eax][20]			// restore EDI
	  mov ebp, [eax][24]			// restore EBP
	  push [eax]
          mov eax, [esp+4]			// restore EAX
          call eax				// Call User's procedure of Application
          add esp, 8
	  popad}
}


